# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-11-13 16:01
from __future__ import unicode_literals

from django.db import migrations, models



class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0001_initial'),
    ]

    operations = [
        # necessary for tests
        migrations.RunSQL("CREATE EXTENSION IF NOT EXISTS citus;"),
        migrations.RunSQL("SELECT * from master_add_node('django-multitenant_worker1_1', 5432);"),
        migrations.RunSQL("SELECT * from master_add_node('django-multitenant_worker2_1', 5432);"),

        # Drop constraints
        migrations.RunSQL("ALTER TABLE tests_aliasedtask DROP CONSTRAINT tests_aliasedtask_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_country DROP CONSTRAINT tests_country_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_manager DROP CONSTRAINT tests_manager_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_project DROP CONSTRAINT tests_project_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_projectmanager DROP CONSTRAINT tests_projectmanager_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_record DROP CONSTRAINT tests_record_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_subtask DROP CONSTRAINT tests_subtask_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_task DROP CONSTRAINT tests_task_pkey CASCADE;"),

        # distribute
        migrations.RunSQL("SELECT create_reference_table('tests_country');"),
        migrations.RunSQL("SELECT create_distributed_table('tests_account', 'id');"),
        migrations.RunSQL("SELECT create_distributed_table('tests_aliasedtask', 'account_id');"),
        migrations.RunSQL("SELECT create_distributed_table('tests_manager', 'account_id');"),
        migrations.RunSQL("SELECT create_distributed_table('tests_organization', 'id');"),
        migrations.RunSQL("SELECT create_distributed_table('tests_project', 'account_id');"),
        migrations.RunSQL("SELECT create_distributed_table('tests_projectmanager', 'account_id');"),
        migrations.RunSQL("SELECT create_distributed_table('tests_record', 'organization_id');"),
        migrations.RunSQL("SELECT create_distributed_table('tests_subtask', 'account_id');"),
        migrations.RunSQL("SELECT create_distributed_table('tests_task', 'account_id');"),

        # Add constraints
        migrations.RunSQL("ALTER TABLE tests_aliasedtask ADD CONSTRAINT tests_aliasedtask_pkey PRIMARY KEY (account_id, id);"),
        migrations.RunSQL("ALTER TABLE tests_country ADD CONSTRAINT tests_country_pkey PRIMARY KEY (id);"),
        migrations.RunSQL("ALTER TABLE tests_project ADD CONSTRAINT tests_project_pkey PRIMARY KEY (account_id, id);"),
        migrations.RunSQL("ALTER TABLE tests_manager ADD CONSTRAINT tests_manager_pkey PRIMARY KEY (account_id, id);"),
        migrations.RunSQL("ALTER TABLE tests_projectmanager ADD CONSTRAINT tests_projectmanager_pkey PRIMARY KEY (account_id, id);"),
        migrations.RunSQL("ALTER TABLE tests_record ADD CONSTRAINT tests_record_pkey PRIMARY KEY (organization_id, id);"),
        migrations.RunSQL("ALTER TABLE tests_subtask ADD CONSTRAINT tests_subtask_pkey PRIMARY KEY (account_id, id);"),
        migrations.RunSQL("ALTER TABLE tests_task ADD CONSTRAINT tests_task_pkey PRIMARY KEY (account_id, id);")
    ]



import logging
from django.apps import apps
from django.db.backends.postgresql.base import (
    DatabaseWrapper as PostgresqlDatabaseWrapper,
    DatabaseSchemaEditor as PostgresqlDatabaseSchemaEditor,
    DatabaseCreation,
    DatabaseFeatures,
    DatabaseOperations,
    DatabaseClient,
    DatabaseIntrospection
)
from django_multitenant.fields import TenantForeignKey

logger = logging.getLogger(__name__)


class DatabaseSchemaEditor(PostgresqlDatabaseSchemaEditor):
    # Override
    def __enter__(self):
        ret = super(DatabaseSchemaEditor, self).__enter__()
        return ret

    # Override
    def _alter_field(self, model, old_field, new_field, old_type, new_type,
                     old_db_params, new_db_params, strict=False):
        if isinstance(old_field, TenantForeignKey):
            if old_field.remote_field and old_field.db_constraint:
                fk_names = self._constraint_names(model, [old_field.column, self.get_tenant_column(model)], foreign_key=True)
                if strict and len(fk_names) != 1:
                    raise ValueError("Found wrong number (%s) of foreign key constraints for %s.%s" % (
                        len(fk_names),
                        model._meta.db_table,
                        old_field.column,
                    ))
                for fk_name in fk_names:
                    logger.info('Deleting foreign key constraint "%s"' % fk_name)
                    sql = self._delete_constraint_sql(self.sql_delete_fk, model, fk_name)
                    for statement in sql.split(';'):
                        self.execute(statement)

        return super(DatabaseSchemaEditor, self)._alter_field(model, old_field, new_field, old_type, new_type,
                     old_db_params, new_db_params, strict)

    # Override
    def _create_fk_sql(self, model, field, suffix):
        if isinstance(field, TenantForeignKey):
            from_table = model._meta.db_table
            from_columns = field.column, self.get_tenant_column(model)
            to_table = field.target_field.model._meta.db_table
            to_columns = field.target_field.column, self.get_tenant_column(field.target_field.model)
            suffix = suffix % {
                "to_table": to_table,
                "to_column": '_'.join(to_columns),
            }

            return self.sql_create_fk % {
                "table": self.quote_name(from_table),
                "name": self.quote_name(self._create_index_name(model, from_columns, suffix=suffix)),
                "column": ', '.join([self.quote_name(from_col) for from_col in from_columns]),
                "to_table": self.quote_name(field.target_field.model._meta.db_table),
                "to_column": ', '.join([self.quote_name(to_col) for to_col in to_columns]),
                "deferrable": self.connection.ops.deferrable_sql(),
            }
        return super(DatabaseSchemaEditor, self)._create_fk_sql(model, field, suffix)

    # Override
    def execute(self, sql, params=()):
        # Hack: Citus will throw the following error if these statements are
        # not executed separately: "ERROR: cannot execute multiple utility events"
        if not params:
            for statement in sql.split(';'):
                if statement:
                    super(DatabaseSchemaEditor, self).execute(statement)
        else:
            super(DatabaseSchemaEditor, self).execute(sql, params)

    @staticmethod
    def get_tenant_column(model):
        app_label = model._meta.app_label
        model_name = model._meta.model_name
        for candidate_model in apps.get_models():
            if (candidate_model._meta.app_label == app_label
                and candidate_model._meta.model_name == model_name):
                return candidate_model.tenant_id
        else:
            # here you can do fallback logic if no model with db_table found
            raise ValueError('Model {}.{} not found!'.format(app_label, model_name))


class DatabaseWrapper(PostgresqlDatabaseWrapper):
    # Override
    SchemaEditorClass = DatabaseSchemaEditor



import logging
from django.db.models.sql.subqueries import DeleteQuery

from .utils import get_tenant_column, get_current_tenant

logger = logging.getLogger(__name__)


def wrap_get_compiler(original_get_compiler):

    def get_compiler(obj, *args, **kwargs):
        from .models import TenantModel
        if issubclass(obj.model, TenantModel):
            current_tenant = get_current_tenant()
            if current_tenant:
                current_tenant_id = getattr(current_tenant,
                                            current_tenant.tenant_id,
                                            None)

                tenant_column = get_tenant_column(obj.model)
                db_table = obj.model._meta.db_table

                # Bad SQL injection
                extra_sql = ['{table}."{column}" = %s'.format(
                    table=db_table, column=tenant_column)]

                extra_params = [current_tenant_id]
                obj.add_extra([],[],extra_sql,extra_params,[],[])
        return original_get_compiler(obj, *args, **kwargs)

    get_compiler._sign = "monkey patch by django-multitenant"
    return get_compiler


def patch_delete_queries_to_include_tenant_ids():
    '''
    Patches the `get_compiler` method of DeleteQuery to include an additional
    filter on tenant_id for any deletions that occur while a tenant is set.

    See https://github.com/citusdata/django-multitenant/issues/15 for a more
    detailed description.
    '''
    if not hasattr(DeleteQuery.get_compiler, "_sign"):
        logger.debug("Patching DeleteQuery.get_compiler()")
        DeleteQuery.get_compiler = wrap_get_compiler(DeleteQuery.get_compiler)
